/*------------------------------------------------------------------------------------------*\
 *   Copyright (C) 2011, ..., 2014 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
\*------------------------------------------------------------------------------------------*/

#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/semaphore.h>
#include <linux/string.h>
#include <linux/sched.h>

#if defined(CONFIG_MACH_ATHEROS) || defined(CONFIG_ATH79)
#include <atheros.h>
#include <atheros_gpio.h>
#include <irq.h>
#endif

#undef INFINEON_CPU
#if defined(CONFIG_VR9) || defined(CONFIG_AR9) || defined(CONFIG_AR10)
#define INFINEON_CPU
#define GPIO_REGISTER_BROKEN
#include <ifx_gpio.h>
#include <ifx_types.h>
#endif

#include "avmnet_debug.h"
#include "avmnet_module.h"
#include "avmnet_config.h"
#include "avmnet_common.h"
#include "athrs_phy_ctrl.h"
#include "avmnet_ar8337.h"

/* Communication with AVM_PA for programming network sessions */
#if defined(CONFIG_AVM_CPMAC_ATH_PPA)
#include <linux/avm_pa.h>
#include <linux/avm_pa_hw.h>
#include <linux/avm_pa_intern.h>

static signed int avmnet_ar8337_add_session(struct avm_pa_session *avm_session);
static signed int avmnet_ar8337_remove_session(struct avm_pa_session *avm_session);

static struct avm_hardware_pa ar8337_hw_pa = {
   .add_session               = avmnet_ar8337_add_session,
   .remove_session            = avmnet_ar8337_remove_session,
   .change_session            = NULL,
   .try_to_accelerate         = NULL,
   .alloc_rx_channel          = NULL,
   .alloc_tx_channel          = NULL,
   .free_rx_channel           = NULL,
   .free_tx_channel           = NULL,
};
#endif /*--- #if defined(CONFIG_AVM_CPMAC_ATH_PPA) ---*/

#define MODULE_NAME "AVMNET_AR8337"
#define AR8337_MAX_PHYS 6

#define SGMII_LINK_WAR_MAX_TRY 10

static struct resource ar8337_gpio = {
    .name  = "ar8337_reset",
    .flags = IORESOURCE_IO,		 
    .start = -1,
    .end   = -1
};

static int read_rmon_all_open(struct inode *inode, struct file *file);
static const struct file_operations read_rmon_all_fops = {
    .open    = read_rmon_all_open,
    .read    = seq_read,
    .llseek  = seq_lseek,
    .release = seq_release,
};

static void reset_switch(avmnet_module_t *this, enum rst_type type);

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void link_work_handler(struct work_struct *data) {
    struct aths17_context *ctx;

    ctx = container_of(to_delayed_work(data), struct aths17_context, link_work);
    ctx->this_module->poll(ctx->this_module);
    enable_irq(ctx->this_module->initdata.swi.irq);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static irqreturn_t ar8337_link_intr(int irq __attribute__((unused)), void *dev) {

    avmnet_module_t *this = (avmnet_module_t *)dev;
    struct aths17_context *ctx = (struct aths17_context *) this->priv;

    AVMNET_DEBUG("[%s] irq %d\n", __func__, this->initdata.swi.irq);
    disable_irq_nosync(this->initdata.swi.irq);

    queue_delayed_work(ctx->workqueue, &(ctx->link_work), 0);

    return IRQ_HANDLED;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_s17_lock(avmnet_module_t *this) {

    struct aths17_context *ctx = (struct aths17_context *)this->priv;

    /*--- printk(KERN_ERR "[%s]\n", __func__); ---*/
    return down_interruptible(&ctx->lock);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void avmnet_s17_unlock(avmnet_module_t *this) {

    struct aths17_context *ctx = (struct aths17_context *)this->priv;

    /*--- printk(KERN_ERR "[%s]\n", __func__); ---*/
    up(&ctx->lock);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned int avmnet_athrs37_reg_read(avmnet_module_t *this, unsigned int s37_addr) {

    avmnet_module_t *parent = this->parent;
    unsigned int addr_temp;
    unsigned int s37_rd_csr_low, s37_rd_csr_high, s37_rd_csr;
    unsigned int data;
    unsigned int phy_address, reg_address;

    if(parent->lock(parent)){
        AVMNET_ERR("[%s] Interrupted while waiting for lock!\n", __func__);
        return 0;
    }

    addr_temp = s37_addr >> 1;
    data = addr_temp >> 8;

    phy_address = 0x18;
    reg_address = 0x00;

    parent->reg_write(parent, phy_address, reg_address, data);

    phy_address = (((addr_temp >> 5) & 0x07) | 0x10);
    reg_address =   (addr_temp & 0x1f);
    s37_rd_csr_low = (uint32_t) parent->reg_read(parent, phy_address, reg_address);

    reg_address++;
    s37_rd_csr_high = (uint32_t) parent->reg_read(parent, phy_address, reg_address);
    s37_rd_csr = (s37_rd_csr_high << 16) | s37_rd_csr_low ;

    parent->unlock(parent);

    return(s37_rd_csr);
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_athrs37_reg_write(avmnet_module_t *this, unsigned int s37_addr, unsigned int s37_write_data) {

    avmnet_module_t *parent = this->parent;
    unsigned int addr_temp;
    unsigned int data;
    unsigned int phy_address, reg_address;

    addr_temp = s37_addr >> 1;
    data = addr_temp >> 8;

    phy_address = 0x18;
    reg_address = 0x0;

    if(parent->lock(parent)){
        AVMNET_ERR("[%s] Interrupted while waiting for lock!\n", __func__);
        return -EINTR;
    }

    parent->reg_write(parent, phy_address, reg_address, data);

    phy_address = (((addr_temp >> 5) & 0x7) | 0x10);
    reg_address = ((addr_temp & 0x1f) | 0x1);

    reg_address--;
    data = s37_write_data  & 0xffff;
    parent->reg_write(parent, phy_address, reg_address, data);

    reg_address++;
    data = (s37_write_data >> 16) & 0xffff;
    parent->reg_write(parent, phy_address, reg_address, data);

    parent->unlock(parent);

    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned int avmnet_s17_rd_phy(avmnet_module_t *this, unsigned int phy_addr, unsigned int reg_addr) {

    avmnet_module_t *parent = this->parent;
    struct aths17_context *ctx;
    unsigned int rddata;

    /*--- AVMNET_DEBUG(KERN_ERR "[%s] Phy 0x%x reg_addr 0x%x\n", __func__, phy_addr, reg_addr); ---*/

    ctx = (struct aths17_context *) this->priv;

    if(down_interruptible(&(ctx->mdio_lock))){
        AVMNET_ERR("[%s] Interrupted while waiting for MDIO mutex\n", __func__);
        return 0xffff;
    }

    /*--- wir können hier gleich per MDIO auf die einzelnen PHYs zugreifen ---*/
    rddata = parent->reg_read(parent, phy_addr, reg_addr);
    rddata = rddata & 0xffff;

    up(&(ctx->mdio_lock));

    return(rddata);
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_s17_wr_phy(avmnet_module_t *this, unsigned int phy_addr, unsigned int reg_addr, unsigned int write_data) {

    avmnet_module_t *parent = this->parent;
    struct aths17_context *ctx;

    /*--- AVMNET_DEBUG(KERN_ERR "[%s] Phy 0x%x reg_addr 0x%x write_data 0x%x\n", __func__, phy_addr, reg_addr, write_data); ---*/

    ctx = (struct aths17_context *) this->priv;

    if(down_interruptible(&(ctx->mdio_lock))){
        AVMNET_ERR("[%s] Interrupted while waiting for MDIO mutex\n", __func__);
        return -EINTR;
    }

    /*--- wir können hier gleich per MDIO auf die einzelnen PHYs zugreifen ---*/
    parent->reg_write(parent, phy_addr, reg_addr, write_data & 0xffff);

    up(&(ctx->mdio_lock));

    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8337_status_poll(avmnet_module_t *this) {
    int              i;
    /*--- uint32_t         IntStatus; ---*/
    /*--- struct aths17_context *ctx = (struct aths17_context *) this->priv; ---*/

    AVMNET_DEBUG("[%s] on %s\n", __func__, this->name);

    for (i = 0; i < this->num_children; i++) {
        /*--- AVMNET_DEBUG("[%s] call %s\n", __func__, this->children[i]->name); ---*/
        this->children[i]->poll(this->children[i]);
    }

#if 0
    if(down_interruptible(&(ctx->mdio_lock))){
        AVMNET_WARN("[%s] Interrupted while waiting for lock!\n", __func__);
        return -EINTR;
    }

    IntStatus = avmnet_athrs37_reg_read(this, S17_GLOBAL_INT1_REG);
    AVMNET_DEBUG("[%s] IntStatus 0x%x\n", __func__, IntStatus);

    avmnet_athrs37_reg_write(this, S17_GLOBAL_INT1_REG, IntStatus); /* clear global link interrupt */

    up(&(ctx->mdio_lock));
#endif

    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8337_set_status(avmnet_module_t *this, avmnet_device_t *device_id, avmnet_linkstatus_t status) {

    AVMNET_TRC("[%s] %s Status %s changed actual %s speed %d\n", 
                __func__, device_id->device_name, this->name, status.Details.link ? "on":"off", status.Details.speed);

    AVMNET_TRC("PORT_STATUS_REG(0)  :%x\n", avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(0) ));
    AVMNET_TRC("PORT_STATUS_REG(1)  :%x\n", avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(1) ));
    AVMNET_TRC("PORT_STATUS_REG(2)  :%x\n", avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(2) ));
    AVMNET_TRC("PORT_STATUS_REG(3)  :%x\n", avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(3) ));
    AVMNET_TRC("PORT_STATUS_REG(4)  :%x\n", avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(4) ));
    AVMNET_TRC("PORT_STATUS_REG(5)  :%x\n", avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(5) ));

    this->parent->set_status(this->parent, device_id, status);

    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8337_setup_interrupt(avmnet_module_t *this, unsigned int on_off) {

    uint32_t    intr_reg_val;
    struct aths17_context *ctx;

    AVMNET_DEBUG("{%s} %sable PHY-Interrupt\n", __func__, on_off ? "en":"dis");

    ctx = (struct aths17_context *) this->priv;

    if (this->initdata.swi.flags & AVMNET_CONFIG_FLAG_IRQ) {
        if(down_interruptible(&(ctx->mdio_lock))){
            AVMNET_WARN("[%s] Interrupted while waiting for lock!\n", __func__);
            return -EINTR;
        }

        if (on_off) {
            intr_reg_val = avmnet_athrs37_reg_read(this, S17_GLOBAL_INT1_REG);   /* clear global link interrupt */
            avmnet_athrs37_reg_write(this, S17_GLOBAL_INT1_REG, intr_reg_val);

            /* Enable global PHY link status interrupt */
            avmnet_athrs37_reg_write(this, S17_GLOBAL_INT1_MASK_REG, 0xFE);
        } else {
            /* disable global PHY link status interrupt */
            avmnet_athrs37_reg_write(this, S17_GLOBAL_INT1_MASK_REG, 0);
        }

        /*--- ath_reg_rmw_clear(ATH_MISC_INT_STATUS, BIT(MISC_BIT_ENET_LINK)); ---*/ /*--- also clear INT_STATUS-Register ---*/

        up(&(ctx->mdio_lock));

        return 0;
    }

    return 1;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void avmnet_ar8337_status_changed(avmnet_module_t *this, avmnet_module_t *caller __attribute__ ((unused))) {

    avmnet_module_t *parent = this->parent;

    AVMNET_DEBUG("[%s] context_name %s\n", __func__, this->name);

    parent->status_changed(parent, this);

}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
#if defined(CONFIG_SGMII_INTERFACE)
#define ATHR_SGMII_FORCED
static unsigned int setup_sgmii(void) {

	uint32_t status = 0, count = 0;

    ath_reg_wr ( ATH_GMAC_MR_AN_CONTROL, 
        ATH_GMAC_MR_AN_CONTROL_POWER_DOWN_SET (1) | 
        ATH_GMAC_MR_AN_CONTROL_PHY_RESET_SET  (1) );

#if defined(ATHR_SGMII_FORCED)
    AVMNET_DEBUG("{%s} SGMII forced Mode\n", __func__);
    ath_reg_wr ( ATH_GMAC_MR_AN_CONTROL, 
        ATH_GMAC_MR_AN_CONTROL_SPEED_SEL1_SET  (1) |
        ATH_GMAC_MR_AN_CONTROL_PHY_RESET_SET   (1) |                                        
        ATH_GMAC_MR_AN_CONTROL_DUPLEX_MODE_SET (1) );

    udelay(10);

    ath_reg_wr ( ATH_GMAC_SGMII_CONFIG, 
        ATH_GMAC_SGMII_CONFIG_MODE_CTRL_SET   (2) |
        ATH_GMAC_SGMII_CONFIG_FORCE_SPEED_SET (1) |                               
        ATH_GMAC_SGMII_CONFIG_SPEED_SET       (2) );
#else
    ath_reg_wr ( ATH_GMAC_SGMII_CONFIG, 
        ATH_GMAC_SGMII_CONFIG_MODE_CTRL_SET (2));

    ath_reg_wr ( ATH_GMAC_MR_AN_CONTROL, 
        ATH_GMAC_MR_AN_CONTROL_AN_ENABLE_SET (1) |
        ATH_GMAC_MR_AN_CONTROL_PHY_RESET_SET (1) );
#endif

    /*------------------------------------------------------------------------------------------*\
	 * SGMII reset sequence suggested by systems team.
    \*------------------------------------------------------------------------------------------*/
	ath_reg_wr ( ATH_GMAC_SGMII_RESET, 
        ATH_GMAC_SGMII_RESET_RX_CLK_N_RESET);

    udelay(10);

	ath_reg_wr ( ATH_GMAC_SGMII_RESET, 
        ATH_GMAC_SGMII_RESET_HW_RX_125M_N_SET (1));

    udelay(10);

	ath_reg_wr ( ATH_GMAC_SGMII_RESET, 
        ATH_GMAC_SGMII_RESET_HW_RX_125M_N_SET (1) |
		ATH_GMAC_SGMII_RESET_RX_125M_N_SET    (1) );

    udelay(10);

	ath_reg_wr ( ATH_GMAC_SGMII_RESET, 
        ATH_GMAC_SGMII_RESET_HW_RX_125M_N_SET (1) |
		ATH_GMAC_SGMII_RESET_TX_125M_N_SET    (1) |
		ATH_GMAC_SGMII_RESET_RX_125M_N_SET    (1) );

    udelay(10);

	ath_reg_wr ( ATH_GMAC_SGMII_RESET, 
        ATH_GMAC_SGMII_RESET_HW_RX_125M_N_SET (1) |
		ATH_GMAC_SGMII_RESET_TX_125M_N_SET    (1) |
		ATH_GMAC_SGMII_RESET_RX_125M_N_SET    (1) |
	    ATH_GMAC_SGMII_RESET_RX_CLK_N_SET     (1) );

    udelay(10);

	ath_reg_wr ( ATH_GMAC_SGMII_RESET, 
        ATH_GMAC_SGMII_RESET_HW_RX_125M_N_SET (1) |
		ATH_GMAC_SGMII_RESET_TX_125M_N_SET    (1) |
		ATH_GMAC_SGMII_RESET_RX_125M_N_SET    (1) |
		ATH_GMAC_SGMII_RESET_RX_CLK_N_SET     (1) |
		ATH_GMAC_SGMII_RESET_TX_CLK_N_SET     (1) );

    mdelay(100);

    /*------------------------------------------------------------------------------------------*\
     * WAR::Across resets SGMII link status goes to weird
	 * state.
	 * if 0xb8070058 (SGMII_DEBUG register) reads other then 0x1f or 0x10
	 * for sure we are in bad  state.
	 * Issue a PHY reset in ATH_GMAC_MR_AN_CONTROL to keep going.
    \*------------------------------------------------------------------------------------------*/
    do {
		ath_reg_rmw_set ( ATH_GMAC_MR_AN_CONTROL, 
            ATH_GMAC_MR_AN_CONTROL_PHY_RESET_SET (1));

		mdelay(200);

		ath_reg_rmw_clear ( ATH_GMAC_MR_AN_CONTROL, 
            ATH_GMAC_MR_AN_CONTROL_PHY_RESET_SET (1));

        mdelay(1000);

		if (count++ == SGMII_LINK_WAR_MAX_TRY) {
			AVMNET_ERR("[%s] Max resets limit reached exiting...\n", __func__);

			return SGMII_LINK_WAR_MAX_TRY;
		}

		status = (ath_reg_rd (ATH_GMAC_SGMII_DEBUG) & 0xff);
    } while (!(status == 0xf || status == 0x10));

	AVMNET_INFO("[%s] Done\n", __func__);

    return 0;
}
#endif /*--- #if defined(CONFIG_SGMII_INTERFACE) ---*/

/*------------------------------------------------------------------------------------------*\
 * bei Honeybee Powerline 1240 ist die CPU an Port2 des AR8334 angeschossen
 * am RGMII-Interface hängt der Powerlinecontroller
\*------------------------------------------------------------------------------------------*/
#if defined(CONFIG_MACH_QCA953x) || defined(CONFIG_SOC_QCA953x)
int avmnet_ar8334_hbee_setup(avmnet_module_t *this) {

    int i, value; 
    unsigned int port;
    struct aths17_context *ctx;

    AVMNET_DEBUG("[%s] %s\n", __func__, this->name);

    ctx = (struct aths17_context *) this->priv;
    
    if(down_interruptible(&(ctx->lock))){
        AVMNET_WARN("[%s] Interrupted while waiting for lock!\n", __func__);
        return -EINTR;
    }

    if (this->initdata.swi.flags & AVMNET_CONFIG_FLAG_RESET) {
        AVMNET_DEBUG("[%s] reset Switch gpio %d\n", __func__, this->initdata.swi.reset);
        reset_switch(this, rst_pulse);
    } else {
        value = avmnet_athrs37_reg_read(this, S17_MASK_CTL_REG);
        AVMNET_DEBUG("{%s} s37 resetting switch... 0x%x ", __func__, value);
        avmnet_athrs37_reg_write(this, S17_MASK_CTL_REG, value | S17_MASK_CTRL_RESET);
    }
    up(&(ctx->lock));

    if(down_interruptible(&(ctx->mdio_lock))){
        AVMNET_WARN("[%s] Interrupted while waiting for lock!\n", __func__);
        return -EINTR;
    }

    i = 0x10;
    while(i--) {
        mdelay(100);
        if( ! (avmnet_athrs37_reg_read(this, S17_MASK_CTL_REG) & S17_MASK_CTRL_RESET))
            break;
    }

    if (i == 0) {
        AVMNET_ERR("[%s] Giving up on hanging Switch %s\n", __func__, this->name);
        return -EFAULT;
    }

    AVMNET_DEBUG("done i %d ControlRegister 0x%x\n", i, avmnet_athrs37_reg_read(this, S17_MASK_CTL_REG));

    /*--- TODO config Interfacetype ---*/
    avmnet_athrs37_reg_write(this, S17_P0_PADCTRL_REG0, S17_MAC0_RGMII_EN);     /* Set RGMII MAC-mode */
    value = avmnet_athrs37_reg_read(this, S17_PWS_REG);
    avmnet_athrs37_reg_write(this, S17_PWS_REG, value | (1<<28) | (1<<7));   /*--- RGMII interface enabled ---*/

    if(this->initdata.swi.flags & AVMNET_CONFIG_FLAG_RX_DELAY){
        AVMNET_DEBUG("{%s} set AVMNET_CONFIG_FLAG_RX_DELAY 0x%x\n", __func__, this->initdata.swi.rx_delay);
        value = avmnet_athrs37_reg_read(this, S17_P0_PADCTRL_REG0);
        value &= ~S17_MAC0_RGMII_RXCLK_MASK;
        value |=  (this->initdata.swi.rx_delay << S17_MAC0_RGMII_RXCLK_SHIFT);
        avmnet_athrs37_reg_write(this, S17_P0_PADCTRL_REG0, value);
        /*--- ACHTUNG: S17_MAC0_RGMII_RXCLK_DELAY befindet sich im S17_P5_PADCTRL_REG1 ---*/
        avmnet_athrs37_reg_write(this, S17_P5_PADCTRL_REG1, S17_MAC0_RGMII_RXCLK_DELAY );
    }
    if(this->initdata.swi.flags & AVMNET_CONFIG_FLAG_TX_DELAY){
        AVMNET_DEBUG("{%s} set AVMNET_CONFIG_FLAG_TX_DELAY 0x%x\n", __func__, this->initdata.swi.tx_delay);
        value = avmnet_athrs37_reg_read(this, S17_P0_PADCTRL_REG0);
        value &= ~S17_MAC0_RGMII_TXCLK_MASK;
        value |= S17_MAC0_RGMII_TXCLK_DELAY | (this->initdata.swi.tx_delay << S17_MAC0_RGMII_TXCLK_SHIFT);
        avmnet_athrs37_reg_write(this, S17_P0_PADCTRL_REG0, value);
    }

    up(&(ctx->mdio_lock));

    this->setup_irq(this, 0);   /*--- irq im Switch ausschalten ---*/

    for (i=0; i< this->num_children; i++) {
        this->children[i]->setup(this->children[i]);    /*--- die internen PHY initialisieren ---*/
    }

    if(down_interruptible(&(ctx->mdio_lock))){
        goto lock_failed;
    }

    value = avmnet_athrs37_reg_read(this, S17_GLOBAL_INT0_REG);   /* clear Interrupts */
    avmnet_athrs37_reg_write(this, S17_GLOBAL_INT0_REG, value);
    value = avmnet_athrs37_reg_read(this, S17_GLOBAL_INT1_REG);   /* clear Interrupts */
    avmnet_athrs37_reg_write(this, S17_GLOBAL_INT1_REG, value);

    AVMNET_DEBUG("{%s} S17_GLOBAL_INTR_REG 0x%x\n", __func__, value);
    AVMNET_DEBUG("{%s} S17_GLOBAL_INTR_REG 0x%x\n", __func__, avmnet_athrs37_reg_read(this, S17_GLOBAL_INT0_REG));

    /*--- clear all VLAN-Entries, disable EEE ---*/
    avmnet_athrs37_reg_write(this, S17_VTU_FUNC1_REG, S17_VTU_FUNC1_VTBUSY | S17_VTU_FUNC1_VT_FUNC_FLUSH);
    avmnet_athrs37_reg_write(this, S17_EEE_CTRL_REG, S17_LPI_DISABLE_ALL);
    avmnet_athrs37_reg_write(this, S17_HDRCTRL_REG, 0);

    value = avmnet_athrs37_reg_read(this, S17_PWR_SEL_REG);
    avmnet_athrs37_reg_write(this, S17_PWR_SEL_REG, value | S17_PWR_SEL_RGMII0_1_8V);

    /*--- value = avmnet_athrs37_reg_read(this, S17_QM_CTRL_REG); ---*/
    /*--- avmnet_athrs37_reg_write(this, S17_QM_CTRL_REG, value | 1 << 6); ---*/    /*--- no packetdrop due flowcontrol ---*/


#if 0       /*--- not connected ---*/
    if (this->initdata.swi.flags & AVMNET_CONFIG_FLAG_IRQ) {
        result = request_irq(ctx->irq, ar8337_link_intr, IRQF_DISABLED | IRQF_TRIGGER_LOW, "ar8337 link", this);
        if (result < 0) {
            AVMNET_ERR("[%s] request irq ar8337 link %d failed\n", __func__, ctx->irq);
            up(&(ctx->mdio_lock));
            return -EINTR;
        }
    }
#endif

    /*--- set packet routing on CPU port ---*/
    /*--- ACHTUNG CPU_PORT_EN muss gesetzt sein, sonst werden Pakete vom Port2 nach Port3 gefiltert ---*/
    value = avmnet_athrs37_reg_read(this, S17_GLOFW_CTRL0_REG);
    value &= ~((3 << 24) | (3 << 22));
    avmnet_athrs37_reg_write(this, S17_GLOFW_CTRL0_REG, value | (2 << 26) | S17_CPU_PORT_EN);
    avmnet_athrs37_reg_write(this, S17_GLOFW_CTRL1_REG, S17_IGMP_JOIN_LEAVE_DPALL | 
                                                        S17_BROAD_DPALL | 
                                                        S17_MULTI_FLOOD_DPALL | 
                                                        S17_UNI_FLOOD_DPALL);

    /*--- enable Atheros-Header 2Bytes ---*/
    avmnet_athrs37_reg_write(this, S17_HDRCTRL_REG, 0);

    /*--- configure Port 0 ---*/
    avmnet_athrs37_reg_write(this, S17_PORT_STATUS_REG(0), S17_PORT_STATUS_AZ_DEFAULT);

    /*--- Port 2 connects to CPU ---*/
    avmnet_athrs37_reg_write(this, S17_PORT_HDRCTRL_REG(2), S17_TXHDR_MODE_ALL | S17_RXHDR_MODE_ALL);
    avmnet_athrs37_reg_write(this, S17_Px_HOL_CTRL0(2), S17_HOL_CTRL0_WAN);
    /*--- avmnet_athrs37_reg_write(this, S17_Px_HOL_CTRL1(2), S17_HOL_CTRL1_DEFAULT); ---*/
    avmnet_athrs37_reg_write(this, S17_Px_HOL_CTRL1(2),  0xCF);
    avmnet_athrs37_reg_write(this, S17_PxLOOKUP_CTRL_REG(2), S17_LOOKUP_CTRL_PORT_VLAN_EN |
                                                             S17_LOOKUP_CTRL_PORTSTATE_FORWARD |
                                                             S17_LOOKUP_CTRL_VLAN_MODE_DISABLE |
                                                             0x7b);

    /* enable external ports for which a config entry exists */
    AVMNET_ERR("{%s} avmnet_hw_config_entry->nr_avm_devices %d\n", __func__, avmnet_hw_config_entry->nr_avm_devices);
    for(i = 0; i < avmnet_hw_config_entry->nr_avm_devices; ++i) {
        port = avmnet_hw_config_entry->avm_devices[i]->vlanID;

        avmnet_athrs37_reg_write(this, S17_Px_HOL_CTRL0(port), S17_HOL_CTRL0_LAN);
        avmnet_athrs37_reg_write(this, S17_Px_HOL_CTRL1(port), S17_HOL_CTRL1_DEFAULT);

        if(port <= AR8337_MAX_PHYS) {
            AVMNET_ERR("{%s} set port %d 0x%X\n", __func__, port, avmnet_athrs37_reg_read(this, S17_PORT_STATUS_REG(port)));
            avmnet_athrs37_reg_write(this, S17_PORT_HDRCTRL_REG(port), 0); /*--- Atheros-Header on Ports ---*/
            avmnet_athrs37_reg_write(this, S17_PxLOOKUP_CTRL_REG(port), S17_LOOKUP_CTRL_PORT_VLAN_EN |
                                                                        S17_LOOKUP_CTRL_PORTSTATE_FORWARD |
                                                                        S17_LOOKUP_CTRL_VLAN_MODE_DISABLE |
                                                                        1 << 2);
        }
    }

    /* clear MAC table */
    avmnet_athrs37_reg_write(this, S17_ATU_FUNC_REG, S17_ATU_FUNC_BUSY | S17_ATU_FUNC_FLUSH_ALL);

    /* Enable MIB counters */
    value = avmnet_athrs37_reg_read(this, S17_MODULE_EN_REG);
    avmnet_athrs37_reg_write(this, S17_MODULE_EN_REG, value | S17_MODULE_MIB_EN | S17_MODULE_ACL_EN);

    /*--- AVMNET_DEBUG("S17_CPU_PORT_REGISTER  :%x\n", avmnet_athrs37_reg_read (this,  S17_CPU_PORT_REG )); ---*/
    for (i = 0; i < 6; i++) 
        AVMNET_ERR("PORT_STATUS_REG(%d)  : 0x%x\n", i, avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(i) ));

    up(&(ctx->mdio_lock));

    return 0;
 
lock_failed:

    if (this->initdata.swi.flags & AVMNET_CONFIG_FLAG_IRQ) {
        free_irq(this->initdata.swi.irq, ar8337_link_intr);
    }

    for (i=0; i< this->num_children; i++) {
         this->children[i]->exit(this->children[i]);
    }

    return -EINTR;
}
#endif

/*------------------------------------------------------------------------------------------*\
 * 7490 Fiber - die Ports im Switch werden hier per VLAN getrennt
\*------------------------------------------------------------------------------------------*/
#if defined(INFINEON_CPU)
int avmnet_ar8334_vlan_setup(avmnet_module_t *this) {
    int    i, value, result;
    unsigned int port;
    struct aths17_context *ctx;

    AVMNET_DEBUG("[%s]\n", __func__);

    ctx = (struct aths17_context *) this->priv;
    
    if(down_interruptible(&(ctx->lock))){
        AVMNET_WARN("[%s] Interrupted while waiting for lock!\n", __func__);
        return -EINTR;
    }

    if (this->initdata.swi.flags & AVMNET_CONFIG_FLAG_RESET) {
        AVMNET_DEBUG("[%s] reset Switch gpio %d\n", __func__, this->initdata.swi.reset);
        reset_switch(this, rst_pulse);
    } else {
        value = avmnet_athrs37_reg_read(this, S17_MASK_CTL_REG);
        AVMNET_DEBUG("{%s} s37 resetting switch... 0x%x ", __func__, value);
        avmnet_athrs37_reg_write(this, S17_MASK_CTL_REG, value | S17_MASK_CTRL_RESET);
    }
    up(&(ctx->lock));

    if(down_interruptible(&(ctx->mdio_lock))){
        AVMNET_WARN("[%s] Interrupted while waiting for lock!\n", __func__);
        return -EINTR;
    }

    i = 0x10;
    while(i--) {
        mdelay(100);
        if( ! (avmnet_athrs37_reg_read(this, S17_MASK_CTL_REG) & S17_MASK_CTRL_RESET))
            break;
    }
    if (i == 0) {
        AVMNET_ERR("[%s] Giving up on hanging Switch %s\n", __func__, this->name);
        return -EFAULT;
    }

    AVMNET_DEBUG("done i %d ControlRegister 0x%x\n", i, avmnet_athrs37_reg_read(this, S17_MASK_CTL_REG));
    
    up(&(ctx->mdio_lock));

    this->setup_irq(this, 0);   /*--- irq im Switch ausschalten ---*/

    for (i=0; i< this->num_children; i++) {
        this->children[i]->setup(this->children[i]);    /*--- die internen PHY initialisieren ---*/
    }

    if(down_interruptible(&(ctx->mdio_lock))){
        goto lock_failed;
    }

    avmnet_athrs37_reg_write(this, S17_PWS_REG, (1<<30));

    value = avmnet_athrs37_reg_read(this, S17_GLOBAL_INT0_REG);   /* clear Interrupts */
    avmnet_athrs37_reg_write(this, S17_GLOBAL_INT0_REG, value);
    value = avmnet_athrs37_reg_read(this, S17_GLOBAL_INT1_REG);   /* clear Interrupts */
    avmnet_athrs37_reg_write(this, S17_GLOBAL_INT1_REG, value);

    AVMNET_DEBUG("{%s} S17_GLOBAL_INTR_REG 0x%x\n", __func__, value);
    AVMNET_DEBUG("{%s} S17_GLOBAL_INTR_REG 0x%x\n", __func__, avmnet_athrs37_reg_read(this, S17_GLOBAL_INT0_REG));

    if (this->initdata.swi.flags & AVMNET_CONFIG_FLAG_IRQ) {
        result = request_irq(this->initdata.swi.irq, ar8337_link_intr, IRQF_DISABLED | IRQF_TRIGGER_LOW, "ar8334 link", this);
        if (result < 0) {
            AVMNET_ERR("[%s] request irq ar8334 link %d failed\n", __func__, this->initdata.swi.irq);
            up(&(ctx->mdio_lock));
            return -EINTR;
        }
    }

    /*--- clear all VLAN-Entries, disable EEE ---*/
    avmnet_athrs37_reg_write(this, S17_VTU_FUNC1_REG, S17_VTU_FUNC1_VTBUSY | S17_VTU_FUNC1_VT_FUNC_FLUSH);
    avmnet_athrs37_reg_write(this, S17_EEE_CTRL_REG, S17_LPI_DISABLE_ALL);
    avmnet_athrs37_reg_write(this, S17_HDRCTRL_REG, 0);

    /*------------------------------------------------------------------------------------------*\
     * config Port0 - CPU-Port
    \*------------------------------------------------------------------------------------------*/
    avmnet_athrs37_reg_write(this, S17_PORT_STATUS_REG(0), S17_PORT_STATUS_AZ_DEFAULT); /*--- Enable flow control on CPU port ---*/ 
    AVMNET_DEBUG("S17_PORT_STATUS_REG 0x%x\n", avmnet_athrs37_reg_read(this, S17_PORT_STATUS_REG(0)));

    value = avmnet_athrs37_reg_read(this, S17_GLOFW_CTRL0_REG); /*--- enable CPU-PORT on Port0 ---*/
    avmnet_athrs37_reg_write(this, S17_GLOFW_CTRL0_REG, value | S17_CPU_PORT_EN | (2 << 24) | (2 << 22));
    avmnet_athrs37_reg_write(this, S17_GLOFW_CTRL1_REG, S17_IGMP_JOIN_LEAVE_DPALL | 
                                                        S17_BROAD_DPALL | 
                                                        S17_MULTI_FLOOD_DPALL | 
                                                        S17_UNI_FLOOD_DPALL);

    /*--- disable Atheros-Header 2Bytes ---*/
    avmnet_athrs37_reg_write(this, S17_PORT_HDRCTRL_REG(0), S17_TXHDR_MODE_NO | S17_RXHDR_MODE_NO);

    avmnet_athrs37_reg_write(this, S17_Px_HOL_CTRL0(0), S17_HOL_CTRL0_WAN);
    avmnet_athrs37_reg_write(this, S17_Px_HOL_CTRL1(0), S17_HOL_CTRL1_DEFAULT);

    /*--- enable VLAN Port0---*/
    avmnet_athrs37_reg_write(this, S17_PxLOOKUP_CTRL_REG(0), S17_LOOKUP_CTRL_LEARN_EN_0 |
                                                             S17_LOOKUP_CTRL_PORTSTATE_FORWARD |
                                                             S17_LOOKUP_CTRL_VLAN_MODE_CHECK |
                                                             0x3E);

    avmnet_athrs37_reg_write(this, S17_PxVLAN_CTRL0_REG(0), 1<<16 | 1); /*--- set default CVID & SVID ---*/
    avmnet_athrs37_reg_write(this, S17_PxVLAN_CTRL1_REG(0), S17_EG_VLAN_MODE_WITH | S17_PxVLAN_CORE_PORT | S17_PxVLAN_PROP_EN_2);

    avmnet_athrs37_reg_write(this, S17_SERVICE_REG, S17_SERVICE_STAG_MODE | 0x8100);
    
    /* enable external ports for which a config entry exists */
    for(i = 0; i < avmnet_hw_config_entry->nr_avm_devices; ++i) {
        port = avmnet_hw_config_entry->avm_devices[i]->vlanID;

        AVMNET_DEBUG("{%s} set port %d 0x%x 0x%X\n", __func__, port, S17_PS_PORT_STATUS_DEFAULT, avmnet_athrs37_reg_read(this, S17_PORT_STATUS_REG(port)));

        // port numbers 1 to AR8337_MAX_PHYS are for SW MACs
        if((port > 0) && (port <= AR8337_MAX_PHYS)) {

            avmnet_athrs37_reg_write(this, S17_Px_HOL_CTRL0(port), S17_HOL_CTRL0_LAN);
            avmnet_athrs37_reg_write(this, S17_Px_HOL_CTRL1(port), S17_HOL_CTRL1_DEFAULT);

            avmnet_athrs37_reg_write(this, S17_PORT_HDRCTRL_REG(port), S17_TXHDR_MODE_NO | S17_RXHDR_MODE_NO);

            avmnet_athrs37_reg_write(this, S17_PxLOOKUP_CTRL_REG(port), 
                                                         S17_LOOKUP_CTRL_LEARN_EN_0 |
                                                         S17_LOOKUP_CTRL_PORTSTATE_FORWARD |
                                                         S17_LOOKUP_CTRL_VLAN_MODE_CHECK |
                                                         (~(1<<port) & 0x3F));

            avmnet_athrs37_reg_write(this, S17_PxVLAN_CTRL0_REG(port), port << 16 | port); /*--- set default CVID & SVID ---*/ 
            avmnet_athrs37_reg_write(this, S17_PxVLAN_CTRL1_REG(port), S17_PxVLAN_PORT_TLS_MODE_2);

            /*--- set VLAN-VID & program VLAN-Table ---*/
            avmnet_athrs37_reg_write(this, S17_VTU_FUNC1_REG, S17_VTU_FUNC1_VTBUSY | S17_VTU_FUNC1_VT_FUNC_LOAD |
                                                              S17_VTU_FUNC1_VID(port));
            if (port == 1) {
                avmnet_athrs37_reg_write(this, S17_VTU_FUNC0_REG, S17_VTU_FUNC0_VALID | S17_VTU_FUNC0_IVL_EN | 
                                                                  S17_VTU_FUNC0_EG_VLANMODE_PORT0(VLANMODE_TAGGED) |
                                                                  S17_VTU_FUNC0_EG_VLANMODE_PORT2(VLANMODE_UNTAGGED) |
                                                                  S17_VTU_FUNC0_EG_VLANMODE_PORT3(VLANMODE_NOMEMBER) |
                                                                  S17_VTU_FUNC0_EG_VLANMODE_PORT6(VLANMODE_NOMEMBER));
            } else {
                avmnet_athrs37_reg_write(this, S17_VTU_FUNC0_REG, S17_VTU_FUNC0_VALID | S17_VTU_FUNC0_IVL_EN | 
                                                                  S17_VTU_FUNC0_EG_VLANMODE_PORT0(VLANMODE_TAGGED) |
                                                                  S17_VTU_FUNC0_EG_VLANMODE_PORT2(VLANMODE_NOMEMBER) |
                                                                  S17_VTU_FUNC0_EG_VLANMODE_PORT3(VLANMODE_UNTAGGED) |
                                                                  S17_VTU_FUNC0_EG_VLANMODE_PORT6(VLANMODE_NOMEMBER));
            }
        }
    }

    /* clear MAC table */
    avmnet_athrs37_reg_write(this, S17_ATU_FUNC_REG, S17_ATU_FUNC_BUSY | S17_ATU_FUNC_FLUSH_ALL);

    /* Enable MIB counters */
    value = avmnet_athrs37_reg_read(this, S17_MODULE_EN_REG);
    avmnet_athrs37_reg_write(this, S17_MODULE_EN_REG, value | S17_MODULE_MIB_EN | S17_MODULE_ACL_EN);

    /*--- AVMNET_DEBUG("S17_CPU_PORT_REGISTER  :%x\n", avmnet_athrs37_reg_read (this,  S17_CPU_PORT_REG )); ---*/
    AVMNET_DEBUG("PORT_STATUS_REG(0)  :%x\n", avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(0) ));
    AVMNET_DEBUG("PORT_STATUS_REG(1)  :%x\n", avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(1) ));
    AVMNET_DEBUG("PORT_STATUS_REG(2)  :%x\n", avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(2) ));
    AVMNET_DEBUG("PORT_STATUS_REG(3)  :%x\n", avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(3) ));
    AVMNET_DEBUG("PORT_STATUS_REG(4)  :%x\n", avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(4) ));
    AVMNET_DEBUG("PORT_STATUS_REG(5)  :%x\n", avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(5) ));

    up(&(ctx->mdio_lock));

    /*--- this->setup_irq(this, 1); ---*/   /*--- hier den IRQ immer einschalten ---*/

    return 0;
 
lock_failed:

    if (this->initdata.swi.flags & AVMNET_CONFIG_FLAG_IRQ) {
        free_irq(this->initdata.swi.irq, ar8337_link_intr);
    }

    for (i=0; i< this->num_children; i++) {
         this->children[i]->exit(this->children[i]);
    }

    return -EINTR;
}
#endif /*--- #if defined(INFINEON_CPU) ---*/

/*------------------------------------------------------------------------------------------*\
 * dieser Code wird nicht benötigt
\*------------------------------------------------------------------------------------------*/
#if 0
int avmnet_ar8337_setup(avmnet_module_t *this) {
    int    i, value, result;
    unsigned int port;
    struct aths17_context *ctx;

    AVMNET_DEBUG("[%s]\n", __func__);

    ctx = (struct aths17_context *) this->priv;
    
    if(down_interruptible(&(ctx->mdio_lock))){
        AVMNET_WARN("[%s] Interrupted while waiting for lock!\n", __func__);
        return -EINTR;
    }

    if (this->initdata.swi.flags & AVMNET_CONFIG_FLAG_RESET) {
        AVMNET_DEBUG("[%s] reset Switch gpio %d\n", __func__, this->initdata.swi.reset);
        reset_switch(this, rst_pulse);
    } else {
        value = avmnet_athrs37_reg_read(this, S17_MASK_CTL_REG);
        AVMNET_DEBUG("{%s} s37 resetting switch... 0x%x ", __func__, value);
        avmnet_athrs37_reg_write(this, S17_MASK_CTL_REG, value | S17_MASK_CTRL_RESET);
    }

    i = 0x10;
    while(i--) {
        mdelay(100);
        if( ! (avmnet_athrs37_reg_read(this, S17_MASK_CTL_REG) & S17_MASK_CTRL_RESET))
            break;
    }
    if (i == 0) {
        AVMNET_ERR("[%s] Giving up on hanging Switch %s\n", __func__, this->name);
        return -EFAULT;
    }

    AVMNET_DEBUG("done i %d ControlRegister 0x%x\n", i, avmnet_athrs37_reg_read(this, S17_MASK_CTL_REG));

    /*--- TODO config Interfacetype ---*/
    avmnet_athrs37_reg_write(this, S17_P0_PADCTRL_REG0, S17_MAC0_SGMII_EN | S17_MAC0_SGMII_DELAY);     /* Set SGMII mode */
    avmnet_athrs37_reg_write(this, S17_SGMII_CTRL_REG, 0xc74164de);      

#if defined(ATHR_SGMII_FORCED)
    avmnet_athrs37_reg_write(this, S17_PWS_REG, 0x602613a0);
#else
    avmnet_athrs37_reg_write(this, S17_PWS_REG, (1<<30));
#endif

    up(&(ctx->mdio_lock));

    this->setup_irq(this, 0);   /*--- irq im Switch ausschalten ---*/

    for (i=0; i< this->num_children; i++) {
        this->children[i]->setup(this->children[i]);    /*--- die internen PHY initialisieren ---*/
    }

    if(down_interruptible(&(ctx->mdio_lock))){
        goto lock_failed;
    }

    /*--- TODO config Interfacetype ---*/
#if defined(CONFIG_SGMII_INTERFACE)
    setup_sgmii();
#endif
 
    value = avmnet_athrs37_reg_read(this, S17_GLOBAL_INT0_REG);   /* clear Interrupts */
    avmnet_athrs37_reg_write(this, S17_GLOBAL_INT0_REG, value);
    value = avmnet_athrs37_reg_read(this, S17_GLOBAL_INT1_REG);   /* clear Interrupts */
    avmnet_athrs37_reg_write(this, S17_GLOBAL_INT1_REG, value);

    AVMNET_DEBUG("{%s} S17_GLOBAL_INTR_REG 0x%x\n", __func__, value);
    AVMNET_DEBUG("{%s} S17_GLOBAL_INTR_REG 0x%x\n", __func__, avmnet_athrs37_reg_read(this, S17_GLOBAL_INT0_REG));

    if (this->initdata.swi.flags & AVMNET_CONFIG_FLAG_IRQ) {
        result = request_irq(ctx->irq, ar8337_link_intr, IRQF_DISABLED | IRQF_TRIGGER_LOW, "ar8337 link", this);
        if (result < 0) {
            AVMNET_ERR("[%s] request irq ar8337 link %d failed\n", __func__, ctx->irq);
            up(&(ctx->mdio_lock));
            return -EINTR;
        }
    }

    /* config CPU Port */ /*--- Enable flow control on CPU port ---*/
    avmnet_athrs37_reg_write(this, S17_PORT_STATUS_REG(0), S17_PORT_STATUS_AZ_DEFAULT);
    AVMNET_DEBUG("S17_PORT_STATUS_REG 0x%x\n", avmnet_athrs37_reg_read(this, S17_PORT_STATUS_REG(0)));

    /*--- enable CPU-PORT on Port0 ---*/
    value = avmnet_athrs37_reg_read(this, S17_GLOFW_CTRL0_REG);
    avmnet_athrs37_reg_write(this, S17_GLOFW_CTRL0_REG, value | S17_CPU_PORT_EN | (2 << 24) | (2 << 22));
    avmnet_athrs37_reg_write(this, S17_GLOFW_CTRL1_REG, 
            S17_IGMP_JOIN_LEAVE_DPALL | S17_BROAD_DPALL | S17_MULTI_FLOOD_DPALL | S17_UNI_FLOOD_DPALL);

    /*--- enable Atheros-Header 2Bytes ---*/
    avmnet_athrs37_reg_write(this, S17_HDRCTRL_REG, 0);
    avmnet_athrs37_reg_write(this, S17_PORT_HDRCTRL_REG(0), S17_TXHDR_MODE_ALL | S17_RXHDR_MODE_ALL);

    avmnet_athrs37_reg_write(this, S17_Px_HOL_CTRL0(0), S17_HOL_CTRL0_WAN);
    avmnet_athrs37_reg_write(this, S17_Px_HOL_CTRL1(0), S17_HOL_CTRL1_DEFAULT);

    avmnet_athrs37_reg_write(this, S17_EEE_CTRL_REG, S17_LPI_DISABLE_ALL);

    avmnet_athrs37_reg_write(this, S17_PxLOOKUP_CTRL_REG(0), 
                                                             S17_LOOKUP_CTRL_PORT_VLAN_EN |
                                                             S17_LOOKUP_CTRL_PORTSTATE_FORWARD |
                                                             S17_LOOKUP_CTRL_VLAN_MODE_DISABLE |
                                                             0x7e);

    /* enable external ports for which a config entry exists */
    for(i = 0; i < avmnet_hw_config_entry->nr_avm_devices; ++i) {
        port = avmnet_hw_config_entry->avm_devices[i]->vlanID;

        AVMNET_DEBUG("{%s} set port %d 0x%x 0x%X\n", __func__, port, S17_PS_PORT_STATUS_DEFAULT, avmnet_athrs37_reg_read(this, S17_PORT_STATUS_REG(port)));

        // port numbers 1 to AR8337_MAX_PHYS are for SW MACs
        if((port > 0) && (port <= AR8337_MAX_PHYS)) {

            /*--- Atheros-Header on Ports ---*/
            avmnet_athrs37_reg_write(this, S17_PORT_HDRCTRL_REG(port), 0);
            if (port == 5) {
                avmnet_athrs37_reg_write(this, S17_Px_HOL_CTRL0(port), S17_HOL_CTRL0_WAN);
                avmnet_athrs37_reg_write(this, S17_Px_HOL_CTRL1(port), S17_HOL_CTRL1_DEFAULT);
            } else {
                avmnet_athrs37_reg_write(this, S17_Px_HOL_CTRL0(port), S17_HOL_CTRL0_LAN);
                avmnet_athrs37_reg_write(this, S17_Px_HOL_CTRL1(port), S17_HOL_CTRL1_DEFAULT);
            }

            avmnet_athrs37_reg_write(this, S17_PxLOOKUP_CTRL_REG(port), 
                                                         S17_LOOKUP_CTRL_PORT_VLAN_EN |
                                                         S17_LOOKUP_CTRL_PORTSTATE_FORWARD |
                                                         S17_LOOKUP_CTRL_VLAN_MODE_DISABLE |
                                                         1);
        }
    }

    /* clear MAC table */
    avmnet_athrs37_reg_write(this, S17_ATU_FUNC_REG, S17_ATU_FUNC_BUSY | S17_ATU_FUNC_FLUSH_ALL);

    /* Enable MIB counters */
    value = avmnet_athrs37_reg_read(this, S17_MODULE_EN_REG);
    avmnet_athrs37_reg_write(this, S17_MODULE_EN_REG, value | S17_MODULE_MIB_EN | S17_MODULE_ACL_EN);

    /*--- AVMNET_DEBUG("S17_CPU_PORT_REGISTER  :%x\n", avmnet_athrs37_reg_read (this,  S17_CPU_PORT_REG )); ---*/
    AVMNET_DEBUG("PORT_STATUS_REG(0)  :%x\n", avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(0) ));
    AVMNET_DEBUG("PORT_STATUS_REG(1)  :%x\n", avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(1) ));
    AVMNET_DEBUG("PORT_STATUS_REG(2)  :%x\n", avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(2) ));
    AVMNET_DEBUG("PORT_STATUS_REG(3)  :%x\n", avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(3) ));
    AVMNET_DEBUG("PORT_STATUS_REG(4)  :%x\n", avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(4) ));
    AVMNET_DEBUG("PORT_STATUS_REG(5)  :%x\n", avmnet_athrs37_reg_read (this,  S17_PORT_STATUS_REG(5) ));

    up(&(ctx->mdio_lock));

    this->setup_irq(this, 1);   /*--- hier den IRQ immer einschalten ---*/

    return 0;
 
lock_failed:

    if (this->initdata.swi.flags & AVMNET_CONFIG_FLAG_IRQ) {
        free_irq(ctx->irq, ar8337_link_intr);
    }

    for (i=0; i< this->num_children; i++) {
         this->children[i]->exit(this->children[i]);
    }

    return -EINTR;
}
#endif

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int avmnet_proc_switch_write(struct file *file __attribute__ ((unused)), const char __user *buf, 
                                    size_t count, 
                                    loff_t *offset __attribute__((unused))) {

    static char cmd[256];
    char *token[3];
    unsigned int number_of_tokens = 0;
    int len;
    unsigned int reg, value;
    avmnet_module_t *this = (avmnet_module_t *) file->private_data;
    struct aths17_context *ctx;

    ctx = (struct aths17_context *) this->priv;

    len = sizeof(cmd) < count ? (sizeof(cmd) - 1) : count;
    len = len - copy_from_user(cmd, buf, len);
    cmd[len] = 0;

    number_of_tokens = avmnet_tokenize(cmd, " ", 3, token);

    if(down_interruptible(&(ctx->mdio_lock))){
        AVMNET_WARN("[%s] Interrupted while waiting for lock!\n", __func__);
        return -EINTR;
    }

    if(   (number_of_tokens == 2)
       && !strncasecmp(token[0], "read", 4)) {
        sscanf(token[1], "%x", &reg);
        value = avmnet_athrs37_reg_read(this, reg);
        AVMNET_SUPPORT("register_read reg %#x = %#x\n", reg, value);
    } else if(   (number_of_tokens == 3)
              && !strncasecmp(token[0], "write", 5)) {
        sscanf(token[1], "%x", &reg);
        sscanf(token[2], "%x", &value);
        avmnet_athrs37_reg_write(this, reg, value);
        AVMNET_SUPPORT("register_write reg %#x = %#x\n", reg, value);
    }

    up(&(ctx->mdio_lock));

    return count;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int avmnet_proc_switch_open(struct inode *inode, struct file *file)
{
	file->private_data = PDE_DATA(inode);

	return 0;
}

static const struct file_operations avmnet_proc_switch_fops = {
	.open = avmnet_proc_switch_open,
	.write = avmnet_proc_switch_write,
};


#if defined(CONFIG_AVM_PA) && defined(CONFIG_AVM_CPMAC_ATH_PPA)
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void avmnet_ar8337_init_pa(void) {
    avm_pa_register_hardware_pa(&ar8337_hw_pa);
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

static inline void read_acl_data_regs(struct aths17_context *ctx,
                                      unsigned int *acl_data_regs,
                                      unsigned int rule_nr,
                                      unsigned int rule_select) {
    unsigned int i, busy_cnt;
    if(down_interruptible(&(ctx->acl_lock))) {
        pr_err("[%s] Interrupted while waiting for lock!\n", __func__);
        return;
    }
    avmnet_athrs37_reg_write(
      ctx->this_module, S17_ACL_FUNC0_REG,
      S17_ACL_FUNC0_REG_ACL_FUNC_INDEX(rule_nr) |
        S17_ACL_FUNC0_REG_ACL_RULE_SELECT(rule_select) |
        S17_ACL_FUNC0_REG_ACL_FUNC(S17_ACL_FUNC0_REG_ACL_FUNC_READ) |
        S17_ACL_FUNC0_REG_ACL_BUSY);
    busy_cnt = 0;
    while(avmnet_athrs37_reg_read(ctx->this_module, S17_ACL_FUNC0_REG) &
          S17_ACL_FUNC0_REG_ACL_BUSY) {
        busy_cnt++;
        schedule();
    }
    if(busy_cnt) pr_debug("acl engine busy for %d cyles\n", busy_cnt);
    for(i = 0; i < 5; i++) {
        acl_data_regs[i] = avmnet_athrs37_reg_read(ctx->this_module,
                                                   S17_ACL_FUNC1_REG + 4 * i);
    }
    up(&ctx->acl_lock);
}

static inline void write_acl_data_regs(struct aths17_context *ctx,
                                       unsigned int *acl_data_regs,
                                       unsigned int rule_nr,
                                       unsigned int rule_select)
{
	unsigned int i;

	if(down_interruptible(&(ctx->acl_lock))) {
		pr_err("[%s] Interrupted while waiting for lock!\n", __func__);
		return;
	}

	for(i = 0; i < 5; i++) {
		avmnet_athrs37_reg_write(ctx->this_module, S17_ACL_FUNC1_REG + 4 * i, acl_data_regs[i]);
	}
	avmnet_athrs37_reg_write( ctx->this_module, S17_ACL_FUNC0_REG,
	                                            S17_ACL_FUNC0_REG_ACL_FUNC_INDEX(rule_nr) | 
                                                S17_ACL_FUNC0_REG_ACL_RULE_SELECT(rule_select) | 
                                                S17_ACL_FUNC0_REG_ACL_FUNC(S17_ACL_FUNC0_REG_ACL_FUNC_WRITE) | 
                                                S17_ACL_FUNC0_REG_ACL_BUSY);
	up(&ctx->acl_lock);
}

#define X XARGS_DECODE
static inline void acl_pattern_mac_decode(struct acl_pattern_mac *dst,
                                          uint32_t *src) {
    memset(dst, 0, sizeof(*dst));
    ACL_MAC_PATTERN_XARGS
}
static inline void acl_mask_mac_decode(struct acl_mask_mac *dst,
                                       uint32_t *src) {
    memset(dst, 0, sizeof(*dst));
    ACL_MAC_MASK_XARGS
}
static inline void acl_pattern_ipv4_decode(struct acl_pattern_ipv4 *dst,
                                           uint32_t *src) {
    memset(dst, 0, sizeof(*dst));
    ACL_IPV4_PATTERN_XARGS
}
static inline void acl_mask_ipv4_decode(struct acl_mask_ipv4 *dst,
                                        uint32_t *src) {
    memset(dst, 0, sizeof(*dst));
    ACL_IPV4_MASK_XARGS
}
static inline void acl_action_decode(struct acl_rule *rule, uint32_t *src) {
    struct acl_action *dst = &rule->action;
    memset(dst, 0, sizeof(*dst));
    ACL_ACTION_XARGS
}
static inline void acl_mask_common_decode(struct acl_mask_common *dst,
                                          uint32_t *src) {
    memset(dst, 0, sizeof(*dst));
    ACL_COMMON_MASK_XARGS
}
#undef X

#define X XARGS_ENCODE
static inline void acl_pattern_mac_encode(struct acl_pattern_mac *src,
                                          uint32_t *dst) {
    ACL_MAC_PATTERN_XARGS
}
static inline void acl_mask_mac_encode(struct acl_mask_mac *src,
                                       uint32_t *dst) {
    ACL_MAC_MASK_XARGS
}
static inline void acl_pattern_ipv4_encode(struct acl_pattern_ipv4 *src,
                                           uint32_t *dst) {
    ACL_IPV4_PATTERN_XARGS
}
static inline void acl_mask_ipv4_encode(struct acl_mask_ipv4 *src,
                                        uint32_t *dst) {
    ACL_IPV4_MASK_XARGS
}
static inline void acl_action_encode(struct acl_rule *rule, uint32_t *dst) {
    struct acl_action *src = &rule->action;
    ACL_ACTION_XARGS
}
#undef X

static inline void acl_pattern_encode(struct acl_rule *rule, uint32_t *regs) {
    switch(rule->type) {
        case ACL_RULE_MAC:
            acl_pattern_mac_encode(&rule->pattern.mac, regs);
            break;
        case ACL_RULE_IPV4:
            acl_pattern_ipv4_encode(&rule->pattern.ipv4, regs);
            break;
        default:
            pr_err("[%s] ACL rule type not implemented\n", __func__);
    }
}

static inline void acl_mask_encode(struct acl_rule *rule, uint32_t *regs) {
    switch(rule->type) {
        case ACL_RULE_MAC:
            acl_mask_mac_encode(&rule->mask.mac, regs);
            break;
        case ACL_RULE_IPV4:
            acl_mask_ipv4_encode(&rule->mask.ipv4, regs);
            break;
        default:
            pr_err("[%s] ACL rule type not implemented\n", __func__);
    }
}

static struct acl_rule *acl_rule_alloc(struct aths17_context *ctx,
                                       gfp_t gfp_flags) {
    unsigned long *usage_map = ctx->acl_rules_map;
    struct semaphore *sema = &ctx->acl_rules_sema;
    unsigned long unused;
    struct acl_rule *new;

    if(down_trylock(sema)) return NULL;

    do {
        unused = find_first_zero_bit(usage_map, AVMNET_AR8337_ACL_RULES);
        if(unused >= AVMNET_AR8337_ACL_RULES) break;
    } while(test_and_set_bit(unused, usage_map));

    new = kzalloc(sizeof(*new), gfp_flags);
    if(new) new->hw_index = unused;
    return new;
}

static void acl_rule_free(struct aths17_context *ctx, struct acl_rule *entry) {
    unsigned long *usage_map = ctx->acl_rules_map;
    struct semaphore *sema = &ctx->acl_rules_sema;

    if(!test_and_clear_bit(entry->hw_index, usage_map)) {
        pr_debug("[%s] handle %d not allocated.\n", __func__, entry->hw_index);
    }

    kfree(entry);

    up(sema);
}

static inline void acl_rule_add(struct aths17_context *ctx,
                                struct acl_rule *acl_rule) {
    uint32_t acl_regs[5];
    struct acl_pattern_ipv4 *ipv4_pattern;
    struct acl_mask_ipv4 *ipv4_mask;

    ipv4_pattern = &acl_rule->pattern.ipv4;
    ipv4_mask = &acl_rule->mask.ipv4;

    /* pattern */
    memset(acl_regs, 0, sizeof(acl_regs));
    acl_pattern_encode(acl_rule, acl_regs);
    write_acl_data_regs(ctx, acl_regs, acl_rule->hw_index,
                        S17_ACL_FUNC0_REG_ACL_RULE_SELECT_RULE);

    /* mask */
    memset(acl_regs, 0, sizeof(acl_regs));
    acl_mask_encode(acl_rule, acl_regs);
    write_acl_data_regs(ctx, acl_regs, acl_rule->hw_index,
                        S17_ACL_FUNC0_REG_ACL_RULE_SELECT_MASK);


    pr_debug("[%s] acl_rule %d rate_sel = %d\n", __func__, acl_rule->hw_index,
             acl_rule->action.rate_sel);
    /* action */
    memset(acl_regs, 0, sizeof(acl_regs));
    acl_action_encode(acl_rule, acl_regs);
    write_acl_data_regs(ctx, acl_regs, acl_rule->hw_index,
                        S17_ACL_FUNC0_REG_ACL_RULE_SELECT_RESULT);
}

static inline void acl_rule_remove(struct aths17_context *ctx,
                                   struct acl_rule *rule) {
    uint32_t acl_regs[5];
    memset(acl_regs, 0, sizeof(acl_regs));
    write_acl_data_regs(ctx, acl_regs, rule->hw_index,
                        S17_ACL_FUNC0_REG_ACL_RULE_SELECT_RULE);
    write_acl_data_regs(ctx, acl_regs, rule->hw_index,
                        S17_ACL_FUNC0_REG_ACL_RULE_SELECT_MASK);
    write_acl_data_regs(ctx, acl_regs, rule->hw_index,
                        S17_ACL_FUNC0_REG_ACL_RULE_SELECT_RESULT);
}

static void bridge_work_setup(struct work_struct *data) {
    struct acl_work *acl_work;
    struct acl_rule *acl_rule;
    struct aths17_context *ctx = NULL;

    acl_work = container_of(data, struct acl_work, work);
    acl_rule = acl_work->acl_rule;
    BUG_ON(!acl_rule);
    ctx = acl_work->ctx;
    BUG_ON(!ctx);

    kfree(acl_work);

    acl_rule_add(ctx, acl_rule);
}

static void bridge_work_destroy(struct work_struct *data) {
    struct acl_work *acl_work;
    struct acl_rule *acl_rule;
    struct aths17_context *ctx = NULL;

    acl_work = container_of(data, struct acl_work, work);
    acl_rule = acl_work->acl_rule;
    BUG_ON(!acl_rule);
    ctx = acl_work->ctx;
    BUG_ON(!ctx);

    kfree(acl_work);

    acl_rule_remove(ctx, acl_rule);
    acl_rule_free(ctx, acl_rule);
}

static const struct acl_mask_mac bridge_acl_mask = {
    .da = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff },
    .sa = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff },
    .rule_type = S17_ACL_RULE_TYPE_MAC,
    .rule_valid = S17_ACL_RULE_VALID_SE,
};

static int bridge_add_pa_session(struct avm_pa_session *avm_pa_session) {
    unsigned int n;
    unsigned int negress = 0;
    struct avm_pa_pid_hwinfo *ingress_hw;
    struct avm_pa_pid_hwinfo *egress_hw;
    struct aths17_context *ctx = NULL;
    struct acl_rule *acl_rule;
    struct acl_pattern_mac *mac_pattern;
    struct acl_action *action;
    struct acl_work *acl_work;

    pr_debug("[%s]\n ", __func__);

    ingress_hw = avm_pa_pid_get_hwinfo(avm_pa_session->ingress_pid_handle);
    ingress_hw = avm_pa_pid_get_hwinfo(avm_pa_session->ingress_pid_handle);

    if((ingress_hw == NULL) ||
       (ingress_hw->flags & AVMNET_DEVICE_IFXPPA_ETH_WAN)) {
        pr_debug(
          "[%s] no bridging acceleration for AVM ATA device [ingress]\n",
          __func__);
        return AVM_PA_TX_ERROR_SESSION;
    }

    ctx =
      (struct aths17_context *)((avmnet_module_t *)(ingress_hw
                                                      ->avmnet_switch_module))
        ->priv;

    acl_rule = acl_rule_alloc(ctx, GFP_ATOMIC);
    if(!acl_rule) {
        pr_debug("[%s] No ACL slots left.\n", __func__);
        return AVM_PA_TX_ERROR_SESSION;
    }

    acl_rule->type = ACL_RULE_MAC;
    acl_rule->mask.mac = bridge_acl_mask;
    mac_pattern = &acl_rule->pattern.mac;
    action = &acl_rule->action;

    mac_pattern->source_port = (1 << ingress_hw->port_number);
    // memcpy(mac_pattern->sa, avm_pa_session->bsession->ethh.h_source,
    // ETH_ALEN);
#if defined(__LITTLE_ENDIAN)
    for(n = 0; n < ETH_ALEN; n++)
        mac_pattern->sa[ETH_ALEN - 1 - n] =
          avm_pa_session->bsession->ethh.h_source[n];
#else
    mac_pattern->sa[0] = avm_pa_session->bsession->ethh.h_source[2];
    mac_pattern->sa[1] = avm_pa_session->bsession->ethh.h_source[3];
    mac_pattern->sa[2] = avm_pa_session->bsession->ethh.h_source[4];
    mac_pattern->sa[3] = avm_pa_session->bsession->ethh.h_source[5];
    mac_pattern->sa[4] = avm_pa_session->bsession->ethh.h_source[0];
    mac_pattern->sa[5] = avm_pa_session->bsession->ethh.h_source[1];
#endif
    // memcpy(mac_pattern->da, avm_pa_session->bsession->ethh.h_dest, ETH_ALEN);
#if defined(__LITTLE_ENDIAN)
    for(n = 0; n < ETH_ALEN; n++)
        mac_pattern->da[ETH_ALEN - 1 - n] =
          avm_pa_session->bsession->ethh.h_dest[n];
#else
    mac_pattern->da[0] = avm_pa_session->bsession->ethh.h_dest[2];
    mac_pattern->da[1] = avm_pa_session->bsession->ethh.h_dest[3];
    mac_pattern->da[2] = avm_pa_session->bsession->ethh.h_dest[4];
    mac_pattern->da[3] = avm_pa_session->bsession->ethh.h_dest[5];
    mac_pattern->da[4] = avm_pa_session->bsession->ethh.h_dest[0];
    mac_pattern->da[5] = avm_pa_session->bsession->ethh.h_dest[1];
#endif

    for(n = 0; n < avm_pa_session->negress; n++) {
        egress_hw = avm_pa_pid_get_hwinfo(avm_pa_session->egress[n].pid_handle);
        if(egress_hw && (egress_hw->flags & (AVMNET_DEVICE_IFXPPA_ETH_LAN))) {
            /* TODO IFX flag?! */
            action->des_port |= 1 << (egress_hw->port_number);
            action->des_port_over_en = 1;
            negress++;
        } else {
            pr_debug("[%s] Ignoring unknown egress device for pid handle %d\n",
                     __func__, avm_pa_session->ingress_pid_handle);
        }
    }
    if(negress == 0) {
        pr_debug("[%s] No egress ports for session found, aborting\n",
                 __func__);
        return AVM_PA_TX_ERROR_SESSION;
    }
    avm_pa_session->hw_session = acl_rule;

    /* setup some work */
    acl_work = kzalloc(sizeof(*acl_work), GFP_ATOMIC);
    BUG_ON(!acl_work);
    acl_work->acl_rule = acl_rule;
    acl_work->ctx = ctx;
    INIT_WORK(&acl_work->work, bridge_work_setup);
    queue_work(ctx->workqueue, &(acl_work->work));

    pr_debug("[%s] added\n ", __func__);
    return AVM_PA_TX_SESSION_ADDED;
}

static signed int
bridge_remove_pa_session(struct avm_pa_session *avm_pa_session) {
    struct acl_rule *acl_rule = avm_pa_session->hw_session;
    struct avm_pa_pid_hwinfo *ingress_hw;
    struct aths17_context *ctx = NULL;
    struct acl_work *acl_work;

    pr_debug("[%s]\n ", __func__);
    ingress_hw = avm_pa_pid_get_hwinfo(avm_pa_session->ingress_pid_handle);
    if(ingress_hw == NULL) {
        pr_debug("[%s] Illegal hw handle for AVM_PA pid %u!\n", __func__,
                       avm_pa_session->ingress_pid_handle);
        return AVM_PA_TX_ERROR_SESSION;
    }
    ctx =
      (struct aths17_context *)((avmnet_module_t *)(ingress_hw
                                                      ->avmnet_switch_module))
        ->priv;

    /* setup some work */
    acl_work = kzalloc(sizeof(*acl_work), GFP_ATOMIC);
    BUG_ON(!acl_work);
    acl_work->acl_rule = acl_rule;
    acl_work->ctx = ctx;
    INIT_WORK(&acl_work->work, bridge_work_destroy);
    queue_work(ctx->workqueue, &(acl_work->work));

    return AVM_PA_TX_OK;
}

static int avmnet_ar8337_remove_session(struct avm_pa_session *avm_pa_session) {
    if(avm_pa_session->bsession) {
        return bridge_remove_pa_session(avm_pa_session);
    }
    return AVM_PA_TX_ERROR_SESSION;
}

__attribute__((unused))
//static void debug_acl_test(struct seq_file *m, void *data) {
static void debug_acl_test(avmnet_module_t *this) {
    unsigned int acl_regs_pattern[5];
    unsigned int acl_regs_mask[5];
    unsigned int acl_regs_action[5];
    struct aths17_context *ctx;
    struct acl_rule *rule;
    //avmnet_module_t *this = data;
    // 06:04:0e:c4:ae:43
    uint8_t hbee_addr[] = { 0x0e, 0xc4, 0xae, 0x43, 0x06, 0x04 };
    uint8_t pc_addr[] = { 0x9f, 0x64, 0xaa, 0x41, 0xa0, 0x36 };
    ctx = (struct aths17_context *)this->priv;

    // works
    ///* occupy all slots */
    //for(i = 0; i < S17_NUMBER_OF_ACL_RULES; i++) {
    //    rule = acl_rule_alloc(ctx, GFP_KERNEL);
    //}
    //if(acl_rule_alloc(ctx, GFP_KERNEL)) {
    //    printk("[%s] expected acl_rule_alloc to fail\n", __func__);
    //}

    rule = acl_rule_alloc(ctx, GFP_KERNEL);
    
    memset(acl_regs_pattern, 0, sizeof(acl_regs_pattern));
    memset(acl_regs_mask, 0, sizeof(acl_regs_mask));
    memset(acl_regs_action, 0, sizeof(acl_regs_action));

    /*
    memcpy(rule->pattern.mac.da, addr1, ETH_ALEN);
    memcpy(rule->pattern.mac.sa, addr2, ETH_ALEN);
    rule->pattern.mac.source_port = 7;
    rule->pattern.mac.rule_res_inv_en = 1;
    
    rule->mask.mac = bridge_acl_mask;
    */
    
    rule->type = ACL_RULE_MAC;
    rule->pattern.mac.source_port = 0xff;
    rule->mask.mac = bridge_acl_mask;
    memcpy(rule->pattern.mac.da, hbee_addr, ETH_ALEN);
    memcpy(rule->pattern.mac.sa, pc_addr, ETH_ALEN);
    rule->action.des_port = 1;
    rule->action.des_port_over_en = true;
    
    
    /*
    rule->type = ACL_RULE_IPV4;
    rule->pattern.ipv4.dport = 0x0;
    rule->pattern.ipv4.dip = (192 << 24) | (168 << 16) | (178 << 8) | (2 << 0);
    rule->pattern.ipv4.sport = 4801;
    //rule->pattern.ipv4.source_port = (1<<2);
    rule->pattern.ipv4.source_port = 0xff;
    rule->mask.ipv4.dip = 0xffffffff;
    rule->mask.ipv4.dport = 0xffff;
    rule->mask.ipv4.sport = 0xffff;
    rule->mask.ipv4.dport_mask_en = false;
    rule->mask.ipv4.sport_mask_en = true;
    rule->mask.ipv4.rule_valid = S17_ACL_RULE_VALID_SE;
    rule->mask.ipv4.rule_type = S17_ACL_RULE_TYPE_IPV4;
    //rule->action.dp_act = 7;
    rule->action.des_port = 1;
    rule->action.des_port_over_en = true;
    */
    
    printk("MAGICMARKER\n");
    acl_pattern_encode(rule, acl_regs_pattern);
    printk("[19:16] 0x%08x\n"
             "[15:12] 0x%08x\n"
             "[11: 8] 0x%08x\n"
             "[ 7: 4] 0x%08x\n"
             "[ 3: 0] 0x%08x\n\n",
             acl_regs_pattern[4],
             acl_regs_pattern[3],
             acl_regs_pattern[2],
             acl_regs_pattern[1],
             acl_regs_pattern[0]);
    acl_mask_encode(rule, acl_regs_mask);
    printk("[19:16] 0x%08x\n"
             "[15:12] 0x%08x\n"
             "[11: 8] 0x%08x\n"
             "[ 7: 4] 0x%08x\n"
             "[ 3: 0] 0x%08x\n\n",
             acl_regs_mask[4],
             acl_regs_mask[3],
             acl_regs_mask[2],
             acl_regs_mask[1],
             acl_regs_mask[0]);
    acl_action_encode(rule, acl_regs_action);
    printk("[19:16] 0x%08x\n"
             "[15:12] 0x%08x\n"
             "[11: 8] 0x%08x\n"
             "[ 7: 4] 0x%08x\n"
             "[ 3: 0] 0x%08x\n\n",
             acl_regs_action[4],
             acl_regs_action[3],
             acl_regs_action[2],
             acl_regs_action[1],
             acl_regs_action[0]);

    write_acl_data_regs(ctx, acl_regs_pattern, rule->hw_index,
                        S17_ACL_FUNC0_REG_ACL_RULE_SELECT_RULE);
    write_acl_data_regs(ctx, acl_regs_mask, rule->hw_index,
                        S17_ACL_FUNC0_REG_ACL_RULE_SELECT_MASK);
    write_acl_data_regs(ctx, acl_regs_action, rule->hw_index,
                        S17_ACL_FUNC0_REG_ACL_RULE_SELECT_RESULT);
}

#if 0
static void debugfs_acl_regs(struct seq_file *m,
                             const char *prefix,
                             unsigned int *acl_regs) {
    int i;
    seq_printf(m, "   %s: \n", prefix);
    for(i = 4; i >= 0; i--) {
        seq_printf(m, "    byte[%2d:%2d] %#.8x\n", (i * 4) + 3, (i * 4),
                   acl_regs[i]);
    }
}

static void debugfs_acl_table(struct seq_file *m, void *data) {
    unsigned int rule;
    unsigned int acl_regs_rule[5];
    unsigned int acl_regs_mask[5];
    unsigned int acl_regs_result[5];
    struct aths17_context *ctx;
    struct acl_mask_common mask;
    avmnet_module_t *this = data;
    ctx = (struct aths17_context *)this->priv;

    for(rule = 0; rule < AVMNET_AR8337_ACL_RULES; rule++) {
        memset(acl_regs_rule, 0, sizeof(acl_regs_rule));
        memset(acl_regs_mask, 0, sizeof(acl_regs_mask));
        memset(acl_regs_result, 0, sizeof(acl_regs_result));

        read_acl_data_regs(ctx, acl_regs_mask, rule,
                           S17_ACL_FUNC0_REG_ACL_RULE_SELECT_MASK);
        acl_mask_common_decode(&mask, acl_regs_mask);
        if(mask.rule_valid) {
            seq_printf(m, "-- valid entry %d: --\n", rule);
            read_acl_data_regs(ctx, acl_regs_rule, rule,
                               S17_ACL_FUNC0_REG_ACL_RULE_SELECT_RULE);
            read_acl_data_regs(ctx, acl_regs_result, rule,
                               S17_ACL_FUNC0_REG_ACL_RULE_SELECT_RESULT);

            seq_printf(m, " type %d (raw output):\n", mask.rule_type);
            debugfs_acl_regs(m, "rule", acl_regs_rule);
            debugfs_acl_regs(m, "mask", acl_regs_mask);
            debugfs_acl_regs(m, "result", acl_regs_result);
        }
    }
}
#endif

/*------------------------------------------------------------------------------------------*\
 * Add AVM_PA session on MAC address layer
 *
 * Return: AVM_PA_TX_ERROR_SESSION or AVM_PA_TX_SESSION_ADDED
\*------------------------------------------------------------------------------------------*/
static int avmnet_ar8337_add_session(struct avm_pa_session *avm_pa_session) {
    if(avm_pa_session->negress == 0) {
        pr_debug("[%s] No egress defined. Refusing session\n", __func__);
        return AVM_PA_TX_ERROR_SESSION;
    }

    if(avm_pa_session->bsession != NULL) {
        return bridge_add_pa_session(avm_pa_session);
    }

    pr_debug("[%s] Session type not supported. Refusing session\n", __func__);
    return AVM_PA_TX_ERROR_SESSION;
}



#endif /*--- #if defined(CONFIG_AVM_PA) && defined(CONFIG_AVM_CPMAC_ATH_PPA) ---*/


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8337_init(avmnet_module_t *this) {

    struct aths17_context *ctx;
    int i, result;

    AVMNET_DEBUG("[%s] Init on module %s called.\n", __func__, this->name);

    ctx = kzalloc(sizeof(struct aths17_context) + (this->num_children * sizeof(avmnet_linkstatus_t)), GFP_KERNEL);
    if(ctx == NULL){
        AVMNET_ERR("[%s] init of avmnet-module %s failed.\n", __func__, this->name);
        return -ENOMEM;
    }

    this->priv = ctx;
    ctx->this_module = this;

    if (this->initdata.swi.flags & AVMNET_CONFIG_FLAG_RESET) {

        if(this->initdata.swi.reset < 100) {
            /*--- Hängt nicht am Shift-Register => GPIO Resourcen anmelden ---*/
            ar8337_gpio.start = this->initdata.swi.reset;
            ar8337_gpio.end = this->initdata.swi.reset;
            if(request_resource(&gpio_resource, &ar8337_gpio)) {
                AVMNET_ERR("[%s] ERROR request resource gpio %d\n", __func__, this->initdata.swi.reset);
                kfree(ctx);
                return -EIO;
            }
        }
    }

    ctx->workqueue = avmnet_get_workqueue();
    if(ctx->workqueue == NULL){
        AVMNET_ERR("[%s] No global workqueue available!\n>", __FUNCTION__);
        kfree(ctx);
        return -EFAULT;
    }
    INIT_DELAYED_WORK(&(ctx->link_work), link_work_handler);

    sema_init(&ctx->lock, 1);
    sema_init(&ctx->mdio_lock, 1);
    sema_init(&ctx->acl_lock, 1);

    memset(ctx->acl_rules_map, 0, sizeof(ctx->acl_rules_map));
    sema_init(&ctx->acl_rules_sema, AVMNET_AR8337_ACL_RULES);

    AVMNET_INFO("{%s} reset %d irq %d\n", __func__, this->initdata.swi.reset, this->initdata.swi.irq);

    for(i = 0; i < this->num_children; ++i){
        result = this->children[i]->init(this->children[i]);
        if(result < 0){
            // handle error
        }
    }

    avmnet_cfg_register_module(this);
    avmnet_cfg_add_seq_procentry(this, "rmon_all", &read_rmon_all_fops);
    avmnet_cfg_add_procentry(this, "switch_reg", &avmnet_proc_switch_fops);

#   if defined(CONFIG_AVM_CPMAC_ATH_PPA)
    avmnet_ar8337_init_pa();
#   endif /*--- #if defined(CONFIG_AVM_PA) ---*/

    return 0;
}
 
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8337_exit(avmnet_module_t *this) {

    int i, result;

    AVMNET_DEBUG("[%s] Exit on module %s called.\n", __func__, this->name);

    for(i = 0; i < this->num_children; ++i){
        result = this->children[i]->exit(this->children[i]);
        if(result != 0){
            // handle error
        }
    }

    if (this->initdata.swi.flags & AVMNET_CONFIG_FLAG_IRQ) {
        free_irq(this->initdata.swi.irq, ar8337_link_intr);
    }

    kfree(this->priv);

    return 0;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int read_rmon_all_show(struct seq_file *seq, void *data __attribute__ ((unused)) ) {
    char *cnt8316names[] = {
              /* 0x00 */    "RxBroad",
              /* 0x04 */    "RxPause",
              /* 0x08 */    "RxMulti",
              /* 0x0c */    "RxFcsErr",
              /* 0x10 */    "RxAlignErr",
              /* 0x14 */    "RxRunt",
              /* 0x18 */    "RxFragment",
              /* 0x1c */    "Rx64Byte",
              /* 0x20 */    "Rx128Byte",
              /* 0x24 */    "Rx256Byte",
              /* 0x28 */    "Rx512Byte",
              /* 0x2c */    "Rx1024Byte",
              /* 0x30 */    "Rx1518Byte",
              /* 0x34 */    "RxMaxByte",
              /* 0x38 */    "RxTooLong",
              /* 0x3c */    "RxGoodByteLow",
              /* 0x40 */    "RxGoodByteHigh",
              /* 0x44 */    "RxBadByteLow",
              /* 0x48 */    "RxBadByteHigh",
              /* 0x4c */    "RxOverflow",
              /* 0x50 */    "Filtered",
              /* 0x54 */    "TxBroad",
              /* 0x58 */    "TxPause",
              /* 0x5c */    "TxMulti",
              /* 0x60 */    "TxUnderRun",
              /* 0x64 */    "Tx64Byte",
              /* 0x68 */    "Tx128Byte",
              /* 0x6c */    "Tx256Byte",
              /* 0x70 */    "Tx512Byte",
              /* 0x74 */    "Tx1024Byte",
              /* 0x78 */    "Tx1518Byte",
              /* 0x7c */    "TxMaxByte",
              /* 0x80 */    "TxOverSize",
              /* 0x84 */    "TxByteLow",
              /* 0x88 */    "TxByteHigh",
              /* 0x8c */    "TxCollision",
              /* 0x90 */    "TxAbortCol",
              /* 0x94 */    "TxMultiCol",
              /* 0x98 */    "TxSingleCol",
              /* 0x9c */    "TxExcDefer",
              /* 0xa0 */    "TxDefer",
              /* 0xa4 */    "TxLateCol"
                           };
    unsigned int i;
    avmnet_module_t *this;
    struct aths17_context *ctx;

    this = (avmnet_module_t *) seq->private;
    ctx = (struct aths17_context *) this->priv;

    if(down_interruptible(&(ctx->mdio_lock))){
        AVMNET_WARN("[%s] Interrupted while waiting for lock!\n", __func__);
        return -EINTR;
    }

    seq_printf(seq, "Counters\n");
    avmnet_athrs37_reg_write(this, S17_MIB_REG, S17_MIB_FUNC_ALL);
    seq_printf(seq, "                    CPU port      Port 1      Port 2      Port 3      Port 4      Port 5\n");
    for(i = 0; i <= 0xa4; i += 4) {
        seq_printf(seq, "  %14s  %10.u  %10.u  %10.u  %10.u  %10.u  %10.u\n", 
                        cnt8316names[i >> 2],
                        avmnet_athrs37_reg_read(this, 0x1000 + i),
                        avmnet_athrs37_reg_read(this, 0x1100 + i),
                        avmnet_athrs37_reg_read(this, 0x1200 + i),
                        avmnet_athrs37_reg_read(this, 0x1300 + i),
                        avmnet_athrs37_reg_read(this, 0x1400 + i),
                        avmnet_athrs37_reg_read(this, 0x1500 + i));
    }
    seq_printf(seq, "\n");

    up(&(ctx->mdio_lock));

    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int read_rmon_all_open(struct inode *inode, struct file *file) {

    return single_open(file, read_rmon_all_show, PDE_DATA(inode));
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
#ifdef INFINEON_CPU
static void reset_switch(avmnet_module_t *this, enum rst_type type)
{
    struct aths17_context *ctx = (struct aths17_context *) this->priv;

    if((this->initdata.swi.flags & AVMNET_CONFIG_FLAG_RESET) == 0){
        AVMNET_ERR("[%s] Reset not configured for module %s\n", __func__, this->name);
        return;
    }

    if(!down_trylock(&(ctx->lock))){
        AVMNET_ERR("[%s] Called without MUTEX held!\n", __func__);
        dump_stack();
        up(&(ctx->lock));
    }

    if(type == rst_on || type == rst_pulse){
        AVMNET_INFOTRC("[%s] PHY %s reset on\n", __func__, this->name);
        ifx_gpio_output_clear(this->initdata.swi.reset, IFX_GPIO_MODULE_EXTPHY_RESET);
        mdelay(10);
    }

    if(type == rst_off || type == rst_pulse){
        AVMNET_INFOTRC("[%s] PHY %s reset off\n", __func__, this->name);
        ifx_gpio_output_set(this->initdata.swi.reset, IFX_GPIO_MODULE_EXTPHY_RESET);
        mdelay(10);
    }
}
#elif defined(CONFIG_MACH_ATHEROS) || defined(CONFIG_ATH79)
static void reset_switch(avmnet_module_t *this, enum rst_type type)
{
    struct aths17_context *ctx = (struct aths17_context *) this->priv;

    if((this->initdata.swi.flags & AVMNET_CONFIG_FLAG_RESET) == 0){
        AVMNET_ERR("[%s] Reset not configured for module %s\n", __func__, this->name);
        return;
    }

    if(!down_trylock(&(ctx->lock))){
        AVMNET_ERR("[%s] Called without MUTEX held!\n", __func__);
        dump_stack();
        up(&(ctx->lock));
    }

    if(type == rst_on || type == rst_pulse){
        ath_avm_gpio_out_bit(this->initdata.swi.reset, 0);
        mdelay(100);
    }

    if(type == rst_off || type == rst_pulse){
        ath_avm_gpio_out_bit(this->initdata.swi.reset, 1);
        mdelay(100);
    }
}
#else
static void reset_switch(avmnet_module_t *this, enum rst_type type __maybe_unused)
{
    AVMNET_ERR("[%s] not implemented yet for module %s!\n", __func__, this->name);
}
#endif // INFINEON_CPU

